/*
 * Copyright (c) 2024 SUSE LLC
 *
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail, you may
 * find current contact information at www.novell.com.
 */


#include <iostream>
#include <set>
#include <boost/range/adaptor/reversed.hpp>

#include "snapper/SystemCmd.h"
#include "snapper/SnapperDefines.h"
#include "snapper/SnapperTmpl.h"

#include "../proxy/proxy.h"
#include "../utils/text.h"

#include "CmdBtrfs.h"
#include "BackupConfig.h"
#include "TheBigThing.h"


namespace snapper
{

    using namespace std;


    const vector<string> EnumInfo<TheBigThing::SourceState>::names({
	"missing", "read-only", "read-write"
    });


    const vector<string> EnumInfo<TheBigThing::TargetState>::names({
	"missing", "valid", "invalid"
    });


    void
    TheBigThing::transfer(const BackupConfig& backup_config, const vector<TheBigThing>& the_big_things)
    {
	// mkdir /test1-backup/1390
	// cp info.xml
	// btrfs send /test1/.snapshots/1390/snapshot/ | btrfs receive /test1-backup/1390

	cout << "transfer " << num << '\n';

	if (source_state == SourceState::MISSING)
	    cerr << "not on source" << '\n';

	switch (backup_config.target_mode)
	{
	    case BackupConfig::TargetMode::LOCAL:
	    {
		SystemCmd::Args cmd_args = { MKDIR_BIN, backup_config.target_path + "/" + to_string(num) };
		SystemCmd cmd(cmd_args);
	    } break;

	    case BackupConfig::TargetMode::SSH_PUSH:
	    {
		SystemCmd::Args cmd_args = { SSH_BIN };
		cmd_args << backup_config.ssh_options_vs();
		cmd_args << MKDIR_BIN " " + quote(backup_config.target_path + "/" + to_string(num));
		SystemCmd cmd(cmd_args);
	    } break;
	}

#if 1

	// There is only one or no parent.

	vector<TheBigThing> parents = detect_parent(the_big_things);

	string cmd_source = BTRFS_BIN " send";
	for (const TheBigThing& parent : parents)
	    cmd_source += " -p " + quote(backup_config.source_path + "/" SNAPSHOTS_NAME "/" +
					 to_string(parent.num) + "/" SNAPSHOT_NAME);
	cmd_source += " " + quote(backup_config.source_path + "/" SNAPSHOTS_NAME "/" + to_string(num) + "/"
				  SNAPSHOT_NAME);

#else

	vector<TheBigThing> clone_sources = detect_clone_sources(the_big_things);

	string cmd_source = BTRFS_BIN " send";
	for (const TheBigThing& clone_source : clone_sources)
	    cmd_source += " -c " + quote(backup_config.source_path + "/" SNAPSHOTS_NAME "/" +
					 to_string(clone_source.num) + "/" SNAPSHOT_NAME);
	cmd_source += " " + quote(backup_config.source_path + "/" SNAPSHOTS_NAME "/" + to_string(num) + "/"
				  SNAPSHOT_NAME);

#endif

	string cmd_target = BTRFS_BIN " receive " + quote(backup_config.target_path + "/" + to_string(num));

	if (true /* debug */)
	{
	    cout << cmd_source << '\n';
	    cout << cmd_target << '\n';
	}

	switch (backup_config.target_mode)
	{
	    case BackupConfig::TargetMode::LOCAL:
	    {
		SystemCmd::Args cmd_args = { SH_BIN, "-c", cmd_source + " | " + cmd_target };
		SystemCmd cmd(cmd_args);
	    } break;

	    case BackupConfig::TargetMode::SSH_PUSH:
	    {
		SystemCmd::Args cmd_args = { SH_BIN, "-c", cmd_source + " | " + SSH_BIN " " +
		    backup_config.ssh_options_s() + " " + quote(cmd_target) };
		SystemCmd cmd(cmd_args);

		if (cmd.retcode() != 0)
		{
		    cerr << "send/receive failed" << '\n';
		    for (const string& s : cmd.get_stderr())
			cout << s << '\n';

		    throw;
		}

	    } break;
	}

	target_state = TargetState::VALID;
    }


    void
    TheBigThing::cleanup(const BackupConfig& backup_config)
    {
	// btrfs subvolume delete /test1-backup/1411/snapshot

	cout << "remove " << num << '\n';

	if (target_state == TargetState::MISSING)
	    cerr << "not on target" << '\n';

	switch (backup_config.target_mode)
	{
	    case BackupConfig::TargetMode::LOCAL:
	    {
		SystemCmd::Args cmd_args = { BTRFS_BIN, "subvolume", "delete", backup_config.target_path + "/" +
		    to_string(num) + "/" SNAPSHOT_NAME };
		SystemCmd cmd(cmd_args);
	    } break;

	    case BackupConfig::TargetMode::SSH_PUSH:
	    {
		SystemCmd::Args cmd_args = { SSH_BIN };
		cmd_args << backup_config.ssh_options_vs();
		cmd_args << BTRFS_BIN " subvolume delete " + quote(backup_config.target_path + "/" +
								   to_string(num) + "/" SNAPSHOT_NAME);
		SystemCmd cmd(cmd_args);
	    } break;
	}

	switch (backup_config.target_mode)
	{
	    case BackupConfig::TargetMode::LOCAL:
	    {
		SystemCmd::Args cmd_args = { RMDIR_BIN, backup_config.target_path + "/" + to_string(num) };
		SystemCmd cmd(cmd_args);
	    } break;

	    case BackupConfig::TargetMode::SSH_PUSH:
	    {
		SystemCmd::Args cmd_args = { SSH_BIN };
		cmd_args << backup_config.ssh_options_vs();
		cmd_args << RMDIR_BIN " " + quote(backup_config.target_path + "/" + to_string(num));
		SystemCmd cmd(cmd_args);
	    } break;
	}

	target_state = TargetState::MISSING;
    }


    vector<TheBigThing>
    TheBigThing::detect_parent(const vector<TheBigThing>& the_big_things) const
    {
	vector<TheBigThing> clone_sources;

	// I could not find documentation about the exact meaning of the parent during a
	// btrfs send. This logic here was developed by trial and error.

	// Find previous snapshots with the same parent UUID. Looks more like a
	// sibling than a parent.

	for (const TheBigThing& tmp1 : boost::adaptors::reverse(the_big_things))
	{
	    if (tmp1.num >= num)
		continue;

	    if (tmp1.source_state != TheBigThing::SourceState::READ_ONLY ||
		tmp1.target_state != TheBigThing::TargetState::VALID)
		continue;

	    if (tmp1.source_parent_uuid == source_parent_uuid)
		return { tmp1 };
	}

	// Find the direct parent.

	for (const TheBigThing& tmp1 : boost::adaptors::reverse(the_big_things))
	{
	    if (tmp1.num >= num)
		continue;

	    if (tmp1.source_state != TheBigThing::SourceState::READ_ONLY ||
		tmp1.target_state != TheBigThing::TargetState::VALID)
		continue;

	    if (tmp1.source_uuid == source_parent_uuid)
		return { tmp1 };
	}

	// Find the direct parent of the direct parent. This case happens after a
	// rollback.

	for (const TheBigThing& tmp1 : boost::adaptors::reverse(the_big_things))
	{
	    if (tmp1.num >= num)
		continue;

	    // The direct parent itself might be read-write and not available on the
	    // target.

	    if (tmp1.source_uuid == source_parent_uuid)
	    {
		for (const TheBigThing& tmp2 : boost::adaptors::reverse(the_big_things))
		{
		    if (tmp2.num >= tmp1.num)
			continue;

		    if (tmp2.source_state != TheBigThing::SourceState::READ_ONLY ||
			tmp2.target_state != TheBigThing::TargetState::VALID)
			continue;

		    if (tmp2.source_uuid == tmp1.source_parent_uuid)
			return { tmp2 };
		}

		break;
	    }
	}

	return {};
    }


    vector<TheBigThing>
    TheBigThing::detect_clone_sources(const vector<TheBigThing>& the_big_things) const
    {
	// with bad clone sources btrfs send reports "ERROR: parent determination failed for xyz"

	vector<TheBigThing> clone_sources;

	set<string> allowed_parent_uuids = { source_parent_uuid };

	for (const TheBigThing& tmp1 : boost::adaptors::reverse(the_big_things))
	{
	    if (tmp1.num >= num)
		continue;

	    /*
	    cout << num << " " << tmp1.source_uuid << " " << tmp1.source_parent_uuid << '\n';

	    if (contains(allowed_parent_uuids, tmp1.source_parent_uuid))
	    {
		if (!contains(allowed_parent_uuids, tmp1.source_parent_uuid))
		    cout << "adding allowed parent uuid " << tmp1.source_parent_uuid << '\n';

		allowed_parent_uuids.insert(tmp1.source_parent_uuid);
	    }
	    */

	    if (tmp1.source_state != TheBigThing::SourceState::READ_ONLY ||
		tmp1.target_state != TheBigThing::TargetState::VALID)
		continue;

	    // Find previous snapshots with the same parent UUID.

	    /*
	    if (tmp1.source_parent_uuid == source_parent_uuid)
		clone_sources.insert(clone_sources.begin(), tmp1);
	    */

	    if (contains(allowed_parent_uuids, tmp1.source_parent_uuid))
		clone_sources.insert(clone_sources.begin(), tmp1);

	    if (clone_sources.size() >= 2)
		break;
	}

	/*
	// Find previous snapshots with the same parent UUID.

	for (const TheBigThing& tmp1 : the_big_things)
	{
	    if (tmp1.source_parent_uuid == source_parent_uuid &&
		tmp1.num < num)
	    {
		if (clone_sources.size() >= 8)
		    clone_sources.erase(clone_sources.begin());

		clone_sources.push_back(tmp1);
	    }
	}
	*/

	// TODO clone source in case of rollback

	return clone_sources;
    }


    bool
    operator<(const TheBigThing& lhs, const TheBigThing& rhs)
    {
	return lhs.num < rhs.num;
    }


    TheBigThings::TheBigThings(const BackupConfig& backup_config, ProxySnappers* snappers, bool verbose)
	: snapper(snappers->getSnapper(backup_config.config)), locker(snapper)
    {
	Shell shell_source;
	Shell shell_target;

	if (backup_config.target_mode == BackupConfig::TargetMode::SSH_PUSH)
	{
	    shell_target.ssh = true;
	    shell_target.ssh_options = backup_config.ssh_options_vs();
	}

	if (backup_config.source_path != snapper->getConfig().getSubvolume())
	{
	    cerr << "path mismatch" << '\n';
	    throw;
	}

	if (verbose)
	    cout << _("Probing source snapshots") << endl;

	// query snapshots on source from snapperd

	const ProxySnapshots& snapshots = snapper->getSnapshots();

	if (verbose)
	    cout << _("Probing extra information for source snapshots") << endl;

	for (const ProxySnapshot& a : snapshots)
	{
	    unsigned int num = a.getNum();

	    if (num == 0)
		continue;

	    // query additional information (uuid, read-only) from btrfs

	    CmdBtrfsSubvolumeShow y(shell_source, backup_config.source_path + "/" SNAPSHOTS_NAME "/" + to_string(num) +
				    "/" SNAPSHOT_NAME);

	    TheBigThing the_big_thing(num);
	    the_big_thing.source_date = a.getDate();
	    the_big_thing.source_state = y.is_read_only() ? TheBigThing::SourceState::READ_ONLY :
		TheBigThing::SourceState::READ_WRITE;
	    the_big_thing.source_uuid = y.get_uuid();
	    the_big_thing.source_parent_uuid = y.get_parent_uuid();

	    the_big_things.push_back(the_big_thing);
	}

	// query snapshots on target from btrfs

	// TODO so far does only work if target path is directly a btrfs. does not work with
	// directories or subvolumes.

	if (verbose)
	    cout << _("Probing target snapshots") << endl;

	CmdBtrfsSubvolumeList x(shell_target, backup_config.target_path);

	if (verbose)
	    cout << _("Probing extra information for target snapshots") << endl;

	for (const CmdBtrfsSubvolumeList::Entry& q : x)
	{
	    // cout << q.path << '\n';

	    string::size_type pos = q.path.find('/');

	    if (pos == string::npos)
	    {
		cerr << "invalid looking subvolume path '" << q.path << "'\n";
		continue;
	    }

	    // TODO

	    string t1 = q.path.substr(0, pos);

	    unsigned int num = stoi(t1);

	    // query additional information (receive-uuid, read-only) from btrfs

	    CmdBtrfsSubvolumeShow y(shell_target, backup_config.target_path + "/" + q.path);

	    bool is_read_only = y.is_read_only();
	    if (!is_read_only)
		cerr << num << " not read-only, maybe interrupted transfer" << '\n';

	    vector<TheBigThing>::iterator it = find(num);
	    if (it != end())
	    {
		// wrong receive-uuid can happen when a snapshots is transferred, then removed
		// and a new one with the same number is generated

		bool correct_uuid = false;

		if (!y.get_received_uuid().empty())
		{
		    if (it->source_uuid == y.get_received_uuid())
			correct_uuid = true;

		    if (!correct_uuid)
		    {
			cout << it->source_uuid << " " << y.get_received_uuid() << '\n';
			cerr << num << " wrong uuid, maybe snapshot number reuse" << '\n';
		    }
		}

		if (correct_uuid && is_read_only)
		    it->target_state = TheBigThing::TargetState::VALID;
		else
		    it->target_state = TheBigThing::TargetState::INVALID;
	    }
	    else
	    {
		TheBigThing the_big_thing(num);

		// cannot check received-uuid so assume valid

		if (is_read_only)
		    the_big_thing.target_state = TheBigThing::TargetState::VALID;
		else
		    the_big_thing.target_state = TheBigThing::TargetState::INVALID;

		it = the_big_things.insert(the_big_things.end(), the_big_thing);
	    }

	    it->target_uuid = y.get_uuid();
	    it->target_parent_uuid = y.get_parent_uuid();
	    it->target_received_uuid = y.get_received_uuid();
	}

	sort(the_big_things.begin(), the_big_things.end());
    }


    void
    TheBigThings::transfer(const BackupConfig& backup_config)
    {
#if 0

	vector<TheBigThing> clone_sources;

	for (TheBigThing& the_big_thing : the_big_things)
	{
	    if (the_big_thing.source_state == TheBigThing::SourceState::READ_ONLY)
	    {
		if (the_big_thing.target_state == TheBigThing::TargetState::INVALID)
		{
		    cout << "removing invalid target " << the_big_thing.num << '\n';
		    the_big_thing.cleanup(backup_config);
		}

		if (the_big_thing.target_state == TheBigThing::TargetState::MISSING)
		{
		    cout << "transferring target " << the_big_thing.num << '\n';
		    the_big_thing.transfer(backup_config, clone_sources);
		}
	    }

	    // TODO better clone source handling?

	    // add clone-source after transfer so that the newly transfer snapshot is also a
	    // clone-source

	    if (the_big_thing.source_state == TheBigThing::SourceState::READ_ONLY &&
		the_big_thing.target_state == TheBigThing::TargetState::VALID)
	    {
		if (clone_sources.size() >= 8)
		    clone_sources.erase(clone_sources.begin());

		clone_sources.push_back(the_big_thing);
	    }
	}

#else

	for (TheBigThing& the_big_thing : the_big_things)
	{
	    if (the_big_thing.source_state == TheBigThing::SourceState::READ_ONLY)
	    {
		if (the_big_thing.target_state == TheBigThing::TargetState::INVALID)
		{
		    cout << "removing invalid target " << the_big_thing.num << '\n';
		    the_big_thing.cleanup(backup_config);
		}

		if (the_big_thing.target_state == TheBigThing::TargetState::MISSING)
		{
		    cout << "transferring target " << the_big_thing.num << '\n';

		    /*
		    vector<TheBigThing> clone_sources;

		    // Find previous snapshots with the same parent UUID.

		    for (const TheBigThing& tmp1 : the_big_things)
		    {
			if (tmp1.source_parent_uuid == the_big_thing.source_parent_uuid &&
			    tmp1.num < the_big_thing.num)
			{
			    if (clone_sources.size() >= 8)
				clone_sources.erase(clone_sources.begin());

			    clone_sources.push_back(tmp1);
			}
		    }

		    // TODO clone source in case of rollback

		    the_big_thing.transfer(backup_config, clone_sources);
		    */

		    the_big_thing.transfer(backup_config, the_big_things);
		}
	    }
	}

#endif
    }


    void
    TheBigThings::cleanup(const BackupConfig& backup_config)
    {
	for (TheBigThing& the_big_thing : the_big_things)
	{
	    if (the_big_thing.target_state == TheBigThing::TargetState::INVALID)
	    {
		cout << "removing invalid target " << the_big_thing.num << '\n';
		the_big_thing.cleanup(backup_config);
	    }

	    if (the_big_thing.source_state == TheBigThing::SourceState::MISSING &&
		the_big_thing.target_state != TheBigThing::TargetState::MISSING)
	    {
		cout << "removing outdated target " << the_big_thing.num << '\n';
		the_big_thing.cleanup(backup_config);
	    }
	}
    }


    vector<TheBigThing>::iterator
    TheBigThings::find(unsigned int num)
    {
	return find_if(begin(), end(), [num](const TheBigThing& the_big_thing) {
	    return the_big_thing.num == num;
	});
    }

}
