/*
 * Copyright (c) 2024 SUSE LLC
 *
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail, you may
 * find current contact information at www.novell.com.
 */


#include <iostream>

#include "snapper/AppUtil.h"
#include "snapper/SnapperDefines.h"

#include "../utils/text.h"
#include "../utils/TableFormatter.h"
#include "../utils/CsvFormatter.h"
#include "../utils/JsonFormatter.h"
#include "../utils/OutputOptions.h"

#include "../proxy/proxy.h"

#include "BackupConfig.h"
#include "GlobalOptions.h"
#include "TheBigThing.h"
#include "cmd.h"


namespace snapper
{

    using namespace std;


    void
    help_list()
    {
	cout << "  " << _("List:") << '\n'
	     << "\t" << _("snbk list") << '\n'
	     << '\n';
    }


    namespace
    {

	enum class Column
	{
	    BACKUP_CONFIG, NUMBER, SOURCE_DATE, SOURCE_STATE, SOURCE_UUID, SOURCE_PARENT_UUID, TARGET_STATE,
	    TARGET_UUID, TARGET_PARENT_UUID, TARGET_RECEIVED_UUID
	};


	Cell
	header_for(Column column)
	{
	    switch (column)
	    {
		case Column::BACKUP_CONFIG:
		    return Cell(_("Backup Config"));

		case Column::NUMBER:
		    return Cell(_("#"), Id::NUMBER, Align::RIGHT);

		case Column::SOURCE_DATE:
		    return Cell(_("Source Date"));

		case Column::SOURCE_STATE:
		    return Cell(_("Source State"));

		case Column::SOURCE_UUID:
		    return Cell(_("(Source) UUID"));

		case Column::SOURCE_PARENT_UUID:
		    return Cell(_("(Source) Parent UUID"));

		case Column::TARGET_STATE:
		    return Cell(_("Target State"));

		case Column::TARGET_UUID:
		    return Cell(_("(Target) UUID"));

		case Column::TARGET_PARENT_UUID:
		    return Cell(_("(Target) Parent UUID"));

		case Column::TARGET_RECEIVED_UUID:
		    return Cell(_("(Target) Received UUID"));
	    }

	    SN_THROW(Exception("invalid column value"));
	    __builtin_unreachable();
	}


	boost::any
	value_for_as_any(const OutputOptions& output_options, Column column, const TheBigThing& value)
	{
	    const string::size_type uuid_cutoff = 14;

	    switch (column)
	    {
		case Column::BACKUP_CONFIG:
		    return string("todo");

		case Column::NUMBER:
		    return value.num;

		case Column::SOURCE_DATE:
		    if (value.source_date == 0)
			return nullptr;
		    return datetime(value.source_date, output_options.utc, output_options.iso);

		case Column::SOURCE_STATE:
		    if (value.source_state == TheBigThing::SourceState::MISSING)
			return nullptr;
		    return toString(value.source_state);

		case Column::SOURCE_UUID:
		    if (value.source_uuid.empty())
			return nullptr;
		    return value.source_uuid.substr(0, uuid_cutoff);

		case Column::SOURCE_PARENT_UUID:
		    if (value.source_parent_uuid.empty())
			return nullptr;
		    return value.source_parent_uuid.substr(0, uuid_cutoff);

		case Column::TARGET_STATE:
		    if (value.target_state == TheBigThing::TargetState::MISSING)
			return nullptr;
		    return toString(value.target_state);

		case Column::TARGET_UUID:
		    if (value.target_uuid.empty())
			return nullptr;
		    return value.target_uuid.substr(0, uuid_cutoff);

		case Column::TARGET_PARENT_UUID:
		    if (value.target_parent_uuid.empty())
			return nullptr;
		    return value.target_parent_uuid.substr(0, uuid_cutoff);

		case Column::TARGET_RECEIVED_UUID:
		    if (value.target_received_uuid.empty())
			return nullptr;
		    return value.target_received_uuid.substr(0, uuid_cutoff);
	    }

	    SN_THROW(Exception("invalid column value"));
	    __builtin_unreachable();
	}


	string
	value_for_as_string(const OutputOptions& output_options, Column column, const TheBigThing& value)
	{
	    return any_to_string(output_options, value_for_as_any(output_options, column, value));
	}


	json_object*
	value_for_as_json(const OutputOptions& output_options, Column column, const TheBigThing& value)
	{
	    return any_to_json(output_options, value_for_as_any(output_options, column, value));
	}


	void
	output_table(const GlobalOptions& global_options, const vector<Column>& columns,
		     const TheBigThings& the_big_things)
	{
	    OutputOptions output_options(global_options.utc(), global_options.iso(), true);

	    TableFormatter formatter(global_options.table_style());

	    for (Column column : columns)
		formatter.header().push_back(header_for(column));

	    for (const TheBigThing& value : the_big_things)
	    {
		vector<string> row;

		for (Column column : columns)
		    row.push_back(value_for_as_string(output_options, column, value));

		formatter.rows().push_back(row);
	    }

	    cout << formatter;
	}


	void
	output_csv(const GlobalOptions& global_options, const vector<Column>& columns,
		   const TheBigThings& the_big_things)
	{
	    OutputOptions output_options(global_options.utc(), global_options.iso(), false);

	    CsvFormatter formatter(global_options.separator(), global_options.headers());

	    for (Column column : columns)
		formatter.header().push_back(toString(column));

	    for (const TheBigThing& value : the_big_things)
	    {
		vector<string> row;

		for (Column column : columns)
		    row.push_back(value_for_as_string(output_options, column, value));

		formatter.rows().push_back(row);
	    }

	    cout << formatter;
	}


	void
	output_json(const GlobalOptions& global_options, const vector<Column>& columns,
		    const TheBigThings& the_big_things)
	{
	    OutputOptions output_options(global_options.utc(), global_options.iso(), false);

	    JsonFormatter formatter;

	    json_object* json_snapshots = json_object_new_array();
	    json_object_object_add(formatter.root(), "snapshots", json_snapshots);

	    for (const TheBigThing& value : the_big_things)
	    {
		json_object* json_snapshot = json_object_new_object();
		json_object_array_add(json_snapshots, json_snapshot);

		for (const Column& column : columns)
		{
		    json_object* tmp =  value_for_as_json(output_options, column, value);
		    if (tmp)
			json_object_object_add(json_snapshot, toString(column).c_str(), tmp);
		}
	    }

	    cout << formatter;
	}

    }


    void
    command_list(GlobalOptions& global_options, GetOpts& get_opts, Hahas& hahas,
		 ProxySnappers* snappers)
    {
	ParsedOpts opts = get_opts.parse("list", GetOpts::no_options);

#if 0
	vector<Column> columns = { Column::NUMBER, Column::SOURCE_DATE, Column::SOURCE_STATE, Column::TARGET_STATE };
#else
	vector<Column> columns = { Column::NUMBER, Column::SOURCE_DATE, Column::SOURCE_STATE, Column::SOURCE_UUID,
	    Column::SOURCE_PARENT_UUID, Column::TARGET_STATE, Column::TARGET_UUID, Column::TARGET_PARENT_UUID,
	    Column::TARGET_RECEIVED_UUID };
#endif

	if (global_options.output_format() != GlobalOptions::OutputFormat::TABLE)
	    columns.insert(columns.begin(), Column::BACKUP_CONFIG);

	if (get_opts.has_args())
	{
	    SN_THROW(OptionsException(_("Command 'list' does not take arguments.")));
	}

	for (const Haha& haha : hahas)
	{
	    // TODO info incase of several backup configs

	    TheBigThings the_big_things(haha.backup_config, snappers, global_options.verbose());

	    switch (global_options.output_format())
	    {
		case GlobalOptions::OutputFormat::TABLE:
		    output_table(global_options, columns, the_big_things);
		    break;

		case GlobalOptions::OutputFormat::CSV:
		    output_csv(global_options, columns, the_big_things);
		    break;

		case GlobalOptions::OutputFormat::JSON:
		    output_json(global_options, columns, the_big_things);
		    break;
	    }
	}
    }


    template <> struct EnumInfo<Column> { static const vector<string> names; };

    const vector<string> EnumInfo<Column>::names({
	"backup-config", "number", "source-date", "source-state", "source-uuid", "source-parent-uuid",
	"target-state", "target-uuid", "target-parent-uuid", "target-received-uuid"
    });

}
